#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <jpeglib.h>

struct Point {
    long r, g, b;
    Point& operator += (const Point& right)
    {
        r += right.r;
        g += right.g;
        b += right.b;
    }
};

struct Image {
    int width, height;
    Point *data;
};

Image *simg[10];

const Point WHITE_POINT = {255, 255, 255};
const int LIMIT = 30*30*3;
const int LIMIT2 = 110;
const int LIMIT_SUM = 200;
const int LIMIT_WIDTH = 4;
const int NUM_COUNT = 5;

bool debuginfo = false;

#define dprintf if(debuginfo) \
    printf

int sqr(int x)
{
    return x*x;
}

int ToGrey(Point *pt)
{
    return (pt->r + pt->b + pt->g) /3;
}

Point* GetPoint(const Image *image, int x, int y)
{
    return &(image->data[x * image->width + y]);
}

bool IsSamePoint(const Point *p1, const Point *p2)
{
    return (sqr(p1->r-p2->r)+sqr(p1->g-p2->g)+sqr(p1->b-p2->b) < LIMIT);
}

void CopyPoint(const Point *p1, Point *p2)
{
    *p2 = *p1;
}

void ClearPoint(Image *image, int x, int y)
{
    Point *p = GetPoint(image, x,y);
    *p = WHITE_POINT;
}

bool LighterThan(const Point *pt, int limit)
{
    return ((pt->r+pt->g+pt->b)/3 > limit);
}

float CompareImage(Image *image, Image *standard)
{
	float count = 0;
	float match = 0;
	for (int i=0; i<image->height; i++)
		for (int j=0; j<image->width; j++)
		{
			Point *PStd = GetPoint(standard, i, j);
			Point *PMy = GetPoint(image, i, j);
			if (LighterThan(PStd, 128))
			{
				// Do nothing
			} else {
				count += 1;
				if (LighterThan(PMy, 128))
					match -= 1;
				else
					match += 1;
			}
		}
//	printf("%d/%d\n", (int)match, (int)count);
	return match / count;
}

int IdentifyImage(Image * image)
{
	float bmatch = 0;
	int imatch = 0;
	for (int i=0; i<10; i++)
	{
		if (CompareImage(image, simg[i]) > bmatch)
		{
			bmatch = CompareImage(image, simg[i]);
			imatch = i;
		}
	}
	return imatch;
}

int load_image(char *filename, Image*& image)
{
    FILE *fin;

    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&cinfo);

    if ((fin = fopen(filename, "rb")) == NULL) {
        fprintf(stderr, "can't open %s\n", filename);
        return ENOENT;
    }

    jpeg_stdio_src(&cinfo, fin);

    jpeg_read_header(&cinfo, TRUE);

    jpeg_start_decompress(&cinfo);

    image = new Image;

    image->data = new Point[cinfo.output_height * cinfo.output_width];

    JSAMPARRAY buffer = new JSAMPROW[1];
    buffer[0] = new JSAMPLE[cinfo.output_width];

    int row = 0;

    while (cinfo.output_scanline < cinfo.output_height)
    {
        jpeg_read_scanlines(&cinfo, buffer, 1);
        for (int i=0; i<cinfo.output_width; i++)
        {
            Point *pt = image->data + row * cinfo.output_width + i;
            pt->r = buffer[0][cinfo.output_components * i * sizeof(JSAMPLE)];
            pt->g = buffer[0][(cinfo.output_components * i + 1) * sizeof(JSAMPLE)];
            pt->b = buffer[0][(cinfo.output_components * i + 2)* sizeof(JSAMPLE)];
/*        memcpy(image->data + row * cinfo.output_width, buffer[0], 
                sizeof(JSAMPLE) * cinfo.output_components * cinfo.output_width);*/
        }
        row++;
    }
    image->width = cinfo.output_width;
    image->height = cinfo.output_height;

    jpeg_finish_decompress(&cinfo);

    jpeg_destroy_decompress(&cinfo);

	return 0;
}

int save_image(char *filename, const Image *image)
{
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
	FILE * ouf;
	if ((ouf = fopen(filename, "wb")) == NULL) {
	    fprintf(stderr, "can't open %s\n", filename);
        return EPERM;
	}
	jpeg_stdio_dest(&cinfo, ouf);

    cinfo.image_width = image->width;
    cinfo.image_height = image->height;
    cinfo.input_components = 1;
    cinfo.in_color_space = JCS_GRAYSCALE;

    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, 100, TRUE);
    
    jpeg_start_compress(&cinfo, TRUE);

	JSAMPROW row_pointer[1];	/* pointer to a single row */
	int row_stride;			/* physical row width in buffer */
    int row = 0;
    
    row_pointer[0] = new JSAMPLE[image->width];

	while (cinfo.next_scanline < cinfo.image_height) {
	    for (int i=0; i<image->width; i++)
        {
            row_pointer[0][i] = ToGrey(GetPoint(image, row, i));
        }
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
        row++;
	}

    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);
    return 0;
}

void PrintImagePart(Image *pimage, int rowStart, int rowEnd, int colStart, int colEnd)
{
    for (int j=rowStart; j<rowEnd+1; j++)
    {
        for (int i=colStart; i<colEnd+1; i++)
        {
            Point *pt = &(pimage->data[j*pimage->width+i]);
            dprintf("%3d", (pt->r+pt->g+pt->b)/3);
        }
        dprintf("\n");
    }
}


void PrintImage(Image *pimage, const char *msg)
{
    dprintf("%s image: \n", msg);
    PrintImagePart(pimage, 0, pimage->height, 0, pimage->width);
}

void PrintNum(Image *pimage, int numCount, int numStart, int numWidth, int rowStart, int rowEnd)
{
    dprintf("Found number %d. Start: %d End: %d Width: %d Height: %d\n", numCount, numStart, numStart+numWidth-1, numWidth, rowEnd-rowStart+1);
    PrintImagePart(pimage, rowStart, rowEnd, numStart, numStart+numWidth-1);
}

Image *SubImage(Image *pimage, int rowStart, int rowEnd, int colStart, int colEnd)
{
	Image* num = new Image;
	num->width = colEnd - colStart + 1;
	num->height = rowEnd - rowStart + 1;
	num->data = new Point[num->width * num->height];
	for (int x=rowStart; x<=rowEnd; x++)
		for (int y=colStart; y<=colEnd; y++)
			num->data[(x - rowStart)*num->width + y - colStart] = *GetPoint(pimage, x, y);
	return num;
}

void SaveSubImage(char *filename, Image *pimage, int rowStart, int rowEnd, int colStart, int colEnd)
{
	PrintImagePart(pimage, rowStart, rowEnd, colStart, colEnd);
	Image *img = SubImage(pimage, rowStart, rowEnd, colStart, colEnd);
	save_image(filename, img);
}

int main(int argc, char *argv[])
{
    Image *pimage;
    if (argc == 1)
    {
        fprintf(stderr, "Usage: %s <infile>\n", argv[0]);
        return 1;
    }

    if (argc == 3)
    {
        debuginfo = true;
    }

    if (int ret = load_image(argv[1], pimage))
    {
        return ret;
    }
	for (int i=0; i<10; i++)
	{
		char filename[20];
		sprintf(filename, "s%d.jpg", i);
		int ret = load_image(filename, simg[i]);
		if (ret)
		{
			fprintf(stderr, "Failed to load %s.\n", filename);
			return ret;
		}
	}
    PrintImage(pimage, "Original"); 

	for (int i=0; i<pimage->height; i++)
	{
		ClearPoint(pimage, i, 0);
		ClearPoint(pimage, i, pimage->width - 1);
	}

	for (int i=0; i<pimage->width; i++)
	{
		ClearPoint(pimage, 0, i);
		ClearPoint(pimage, pimage->height - 1, i);
	}

	for (int i=0; i<pimage->height; i++)
	{
		if ((i==0)||(i==pimage->height-1)) continue;
		for (int j=0; j<pimage->width; j++)
		{
			if ((j==0)||(j==pimage->width-1)) continue;
			Point* pt = GetPoint(pimage, i, j);
			int ct = 0;
			if (LighterThan(GetPoint(pimage, i-1, j), 128))
				ct++;
			if (LighterThan(GetPoint(pimage, i+1, j), 128))
				ct++;
			if (LighterThan(GetPoint(pimage, i, j-1), 128))
				ct++;
			if (LighterThan(GetPoint(pimage, i, j+1), 128))
				ct++;
			if (ct >= 3)
				ClearPoint(pimage, i, j);
		}
	}
    PrintImage(pimage, "Processed"); 

	Image *n1 = SubImage(pimage, 2, 16, 3, 11);
	printf("%d", IdentifyImage(n1));
	save_image("n1.jpg", n1);
	Image *n2 = SubImage(pimage, 2, 16, 14, 22);
	printf("%d", IdentifyImage(n2));
	save_image("n2.jpg", n2);
	Image *n3 = SubImage(pimage, 2, 16, 25, 33);
	printf("%d", IdentifyImage(n3));
	save_image("n3.jpg", n3);
	Image *n4 = SubImage(pimage, 2, 16, 36, 44);
	printf("%d", IdentifyImage(n4));
	save_image("n4.jpg", n4);

    return 0;
}
